@@ -4,7 +4,8 @@ require 5.004;
 
 use English;
 use strict;
-use vars qw($VERSION @ISA %Handler_Setters %Encoding_Table @Encoding_Path);
+use vars qw($VERSION @ISA %Handler_Setters %Encoding_Table @Encoding_Path
+	    $have_File_Spec $Attdef_Flag);
 use Carp;
 
 use IO::Handle;
@@ -12,10 +13,24 @@ use IO::Handle;
 require DynaLoader;
 
 @ISA = qw(DynaLoader);
-$VERSION = "2.18" ;
+$VERSION = "2.20" ;
+$Attdef_Flag = 1 << 30;
+
+my $namespace_mask = $Attdef_Flag - 1;
+
+$have_File_Spec = do 'File/Spec.pm';
 
 %Encoding_Table = ();
-@Encoding_Path = (grep(-d $_, map($_ . '/XML/Parser/Encodings', @INC)), '.');
+if ($have_File_Spec) {
+  @Encoding_Path = (grep(-d $_,
+			 map(File::Spec->catdir($_, qw(XML Parser Encodings)),
+			     @INC)),
+		    File::Spec->curdir);
+}
+else {
+  @Encoding_Path = (grep(-d $_, map($_ . '/XML/Parser/Encodings', @INC)), '.');
+}
+  
 
 bootstrap XML::Parser::Expat $VERSION;
 
@@ -25,6 +40,8 @@ bootstrap XML::Parser::Expat $VERSION;
 		    Char  => \&SetCharacterDataHandler,
 		    Proc  => \&SetProcessingInstructionHandler,
 		    Comment => \&SetCommentHandler,
+		    CdataStart => \&SetStartCdataHandler,
+		    CdataEnd   => \&SetEndCdataHandler,
 		    Default => \&SetDefaultHandler,
 		    Unparsed => \&SetUnparsedEntityDeclHandler,
 		    Notation => \&SetNotationDeclHandler,
@@ -57,7 +74,9 @@ sub load_encoding {
   $file .= '.enc' unless $file =~ /\.enc$/;
   unless ($file =~ m!^/!) {
     foreach (@Encoding_Path) {
-      my $tmp = "$_/$file";
+      my $tmp = ($have_File_Spec
+		 ? File::Spec->catfile($_, $file)
+		 : "$_/$file");
       if (-e $tmp) {
 	$file = $tmp;
 	last;
@@ -133,6 +152,11 @@ sub recognized_string {
   RecognizedString($self->{Parser});
 }
 
+sub original_string {
+  my $self = shift;
+  OriginalString($self->{Parser});
+}
+
 sub current_line {
   GetCurrentLineNumber($_[0]->{Parser});
 }
@@ -183,17 +207,24 @@ sub depth {
   int(@{$self->{Context}});
 }
 
+sub element_index {
+  my ($self) = @_;
+
+  ElementIndex($self->{Parser});
+}
+
 sub namespace {
   my ($self, $name) = @_;
   local($WARNING) = 0;
-  $self->{Namespace_List}->[$name];
+  $self->{Namespace_List}->[int($name) & $namespace_mask];
 }
 
 sub eq_name {
   my ($self, $nm1, $nm2) = @_;
   local($WARNING) = 0;
 
-  $nm1 == $nm2 and $nm1 eq $nm2;
+  ($namespace_mask & int($nm1)) == ($namespace_mask & int($nm2))
+    and $nm1 eq $nm2;
 }
 
 sub generate_ns_name {
@@ -205,6 +236,19 @@ sub generate_ns_name {
       : $name;
 }
 
+sub is_defaulted {
+  my ($self, $attname) = @_;
+  local($WARNING) = 0;
+
+  return (int($attname) & $Attdef_Flag) == $Attdef_Flag;
+}
+
+sub finish {
+  my ($self) = @_;
+  my $parser = $self->{Parser};
+  UnsetAllHandlers($parser);
+}
+
 sub position_in_context {
   my ($self, $lines) = @_;
   my $parser = $self->{Parser};
@@ -226,6 +270,11 @@ sub position_in_context {
   $ret;
 }
 
+sub release {
+  my $self = shift;
+  ParserRelease($self->{Parser});
+}
+
 sub DESTROY {
   my $self = shift;
   ParserFree($self->{Parser});
@@ -442,6 +491,14 @@ This event is generated when a processing instruction is recognized.
 
 This event is generated when a comment is recognized.
 
+=item * CdataStart	(Parser)
+
+This is called at the start of a CDATA section.
+
+=item * CdataEnd	(Parser)
+
+This is called at the end of a CDATA section.
+
 =item * Default		(Parser, String)
 
 This is called for any characters that don't have a registered handler.
@@ -561,6 +618,12 @@ Returns the string from the document that was recognized in order to call
 the current handler. For instance, when called from a start handler, it
 will give us the the start-tag string. The string is encoded in UTF-8.
 
+=item original_string
+
+Returns the verbatim string from the document that was recognized in
+order to call the current handler. The string is in the original document
+encoding.
+
 =item default_current
 
 When called from a handler, causes the sequence of characters that generated
@@ -624,6 +687,14 @@ method to create the NAME argument.
 
 Returns the size of the context list.
 
+=item element_index
+
+Returns an integer that is the depth-first visit order of the current
+element. This will be zero outside of the root element. Note that
+in start and end handlers, you're outside of the element that the
+corresponding tags delimit. For example, this would return 0 when
+called from the start handler for the root element start tag.
+
 =item position_in_context(LINES)
 
 Returns a string that shows the current parse position. LINES should be
@@ -650,6 +721,17 @@ This method is deprecated in favor of the parse method.
 Parses the XML document in the given file. Will die if parsestring or
 parsefile has been called previously for this instance.
 
+= item is_defaulted(ATTNAME)
+
+When passed an attribute name received through the start handler, returns
+true if the attribute was supplied as a default.
+
+=item finish
+
+Unsets all handlers (including internal ones that set context), but expat
+continues parsing to the end of the document or until it finds an error.
+It should finish up a lot faster than with the handlers set.
+
 =back
 
 =head1 FUNCTIONS
@@ -678,7 +760,7 @@ explicitly load an encoding not contained in the @Encoding_Path list.
 
 Larry Wall <F<larry@wall.org>> wrote version 1.0.
 
-Clark Cooper <F<coopercl@sch.ge.com>> picked up support, changed the API
+Clark Cooper <F<coopercc@netheaven.com>> picked up support, changed the API
 for this version (2.x), provided documentation, and added some standard
 package features.
 
