@@ -18,11 +18,8 @@ James Clark. All Rights Reserved.
 Contributor(s):
 */
 
-#include <stdlib.h>
-#include <string.h>
-#include <stddef.h>
-
 #include "xmldef.h"
+#include "xmlparse.h"
 
 #ifdef XML_UNICODE
 #define XML_ENCODE_MAX XML_UTF16_ENCODE_MAX
@@ -43,7 +40,7 @@ typedef char ICHAR;
 #endif
 
 
-#ifndef XMLNS
+#ifndef XML_NS
 
 #define XmlInitEncodingNS XmlInitEncoding
 #define XmlInitUnknownEncodingNS XmlInitUnknownEncoding
@@ -63,7 +60,6 @@ typedef char ICHAR;
 /* Round up n to be a multiple of sz, where sz is a power of 2. */
 #define ROUND_UP(n, sz) (((n) + ((sz) - 1)) & ~((sz) - 1))
 
-#include "xmlparse.h"
 #include "xmltok.h"
 #include "xmlrole.h"
 #include "hashtable.h"
@@ -76,6 +72,10 @@ typedef char ICHAR;
 
 #define EXPAND_SPARE 24
 
+/* On macintosh, '\n' may get translated to 0xD instead of 0xA */
+#define CR '\015'
+#define LF '\012'
+
 typedef struct binding {
   struct prefix *prefix;
   struct binding *nextTagBinding;
@@ -268,10 +268,13 @@ typedef struct {
   XML_CharacterDataHandler characterDataHandler;
   XML_ProcessingInstructionHandler processingInstructionHandler;
   XML_CommentHandler commentHandler;
+  XML_StartCdataSectionHandler startCdataSectionHandler;
+  XML_EndCdataSectionHandler endCdataSectionHandler;
   XML_DefaultHandler defaultHandler;
   XML_UnparsedEntityDeclHandler unparsedEntityDeclHandler;
   XML_NotationDeclHandler notationDeclHandler;
   XML_ExternalEntityRefHandler externalEntityRefHandler;
+  void *externalEntityRefHandlerArg;
   XML_UnknownEncodingHandler unknownEncodingHandler;
   const ENCODING *encoding;
   INIT_ENCODING initEncoding;
@@ -319,10 +322,13 @@ typedef struct {
 #define characterDataHandler (((Parser *)parser)->characterDataHandler)
 #define processingInstructionHandler (((Parser *)parser)->processingInstructionHandler)
 #define commentHandler (((Parser *)parser)->commentHandler)
+#define startCdataSectionHandler (((Parser *)parser)->startCdataSectionHandler)
+#define endCdataSectionHandler (((Parser *)parser)->endCdataSectionHandler)
 #define defaultHandler (((Parser *)parser)->defaultHandler)
 #define unparsedEntityDeclHandler (((Parser *)parser)->unparsedEntityDeclHandler)
 #define notationDeclHandler (((Parser *)parser)->notationDeclHandler)
 #define externalEntityRefHandler (((Parser *)parser)->externalEntityRefHandler)
+#define externalEntityRefHandlerArg (((Parser *)parser)->externalEntityRefHandlerArg)
 #define unknownEncodingHandler (((Parser *)parser)->unknownEncodingHandler)
 #define encoding (((Parser *)parser)->encoding)
 #define initEncoding (((Parser *)parser)->initEncoding)
@@ -371,6 +377,15 @@ typedef struct {
 #define hadExternalDoctype (((Parser *)parser)->hadExternalDoctype)
 #define namespaceSeparator (((Parser *)parser)->namespaceSeparator)
 
+#ifdef _MSC_VER
+#ifdef _DEBUG
+Parser *asParser(XML_Parser parser)
+{
+  return parser;
+}
+#endif
+#endif
+
 XML_Parser XML_ParserCreate(const XML_Char *encodingName)
 {
   XML_Parser parser = malloc(sizeof(Parser));
@@ -385,10 +400,13 @@ XML_Parser XML_ParserCreate(const XML_Char *encodingName)
   characterDataHandler = 0;
   processingInstructionHandler = 0;
   commentHandler = 0;
+  startCdataSectionHandler = 0;
+  endCdataSectionHandler = 0;
   defaultHandler = 0;
   unparsedEntityDeclHandler = 0;
   notationDeclHandler = 0;
   externalEntityRefHandler = 0;
+  externalEntityRefHandlerArg = parser;
   unknownEncodingHandler = 0;
   buffer = 0;
   bufferPtr = 0;
@@ -466,6 +484,18 @@ XML_Parser XML_ParserCreateNS(const XML_Char *encodingName, XML_Char nsSep)
   return parser;
 }
 
+int XML_SetEncoding(XML_Parser parser, const XML_Char *encodingName)
+{
+  if (!encodingName)
+    protocolEncodingName = 0;
+  else {
+    protocolEncodingName = poolCopyString(&tempPool, encodingName);
+    if (!protocolEncodingName)
+      return 0;
+  }
+  return 1;
+}
+
 XML_Parser XML_ExternalEntityParserCreate(XML_Parser oldParser,
 					  const XML_Char *context,
 					  const XML_Char *encodingName)
@@ -477,12 +507,15 @@ XML_Parser XML_ExternalEntityParserCreate(XML_Parser oldParser,
   XML_CharacterDataHandler oldCharacterDataHandler = characterDataHandler;
   XML_ProcessingInstructionHandler oldProcessingInstructionHandler = processingInstructionHandler;
   XML_CommentHandler oldCommentHandler = commentHandler;
+  XML_StartCdataSectionHandler oldStartCdataSectionHandler = startCdataSectionHandler;
+  XML_EndCdataSectionHandler oldEndCdataSectionHandler = endCdataSectionHandler;
   XML_DefaultHandler oldDefaultHandler = defaultHandler;
   XML_ExternalEntityRefHandler oldExternalEntityRefHandler = externalEntityRefHandler;
   XML_UnknownEncodingHandler oldUnknownEncodingHandler = unknownEncodingHandler;
   void *oldUserData = userData;
   void *oldHandlerArg = handlerArg;
   int oldDefaultExpandInternalEntities = defaultExpandInternalEntities;
+  void *oldExternalEntityRefHandlerArg = externalEntityRefHandlerArg;
  
   parser = (ns
             ? XML_ParserCreateNS(encodingName, namespaceSeparator)
@@ -494,6 +527,8 @@ XML_Parser XML_ExternalEntityParserCreate(XML_Parser oldParser,
   characterDataHandler = oldCharacterDataHandler;
   processingInstructionHandler = oldProcessingInstructionHandler;
   commentHandler = oldCommentHandler;
+  startCdataSectionHandler = oldStartCdataSectionHandler;
+  endCdataSectionHandler = oldEndCdataSectionHandler;
   defaultHandler = oldDefaultHandler;
   externalEntityRefHandler = oldExternalEntityRefHandler;
   unknownEncodingHandler = oldUnknownEncodingHandler;
@@ -502,6 +537,8 @@ XML_Parser XML_ExternalEntityParserCreate(XML_Parser oldParser,
     handlerArg = userData;
   else
     handlerArg = parser;
+  if (oldExternalEntityRefHandlerArg != oldParser)
+    externalEntityRefHandlerArg = oldExternalEntityRefHandlerArg;
   defaultExpandInternalEntities = oldDefaultExpandInternalEntities;
   if (!dtdCopy(&dtd, oldDtd) || !setContext(parser, context)) {
     XML_ParserFree(parser);
@@ -612,6 +649,14 @@ void XML_SetCommentHandler(XML_Parser parser,
   commentHandler = handler;
 }
 
+void XML_SetCdataSectionHandler(XML_Parser parser,
+				XML_StartCdataSectionHandler start,
+			        XML_EndCdataSectionHandler end)
+{
+  startCdataSectionHandler = start;
+  endCdataSectionHandler = end;
+}
+
 void XML_SetDefaultHandler(XML_Parser parser,
 			   XML_DefaultHandler handler)
 {
@@ -644,6 +689,14 @@ void XML_SetExternalEntityRefHandler(XML_Parser parser,
   externalEntityRefHandler = handler;
 }
 
+void XML_SetExternalEntityRefHandlerArg(XML_Parser parser, void *arg)
+{
+  if (arg)
+    externalEntityRefHandlerArg = arg;
+  else
+    externalEntityRefHandlerArg = parser;
+}
+
 void XML_SetUnknownEncodingHandler(XML_Parser parser,
 				   XML_UnknownEncodingHandler handler,
 				   void *data)
@@ -771,6 +824,13 @@ long XML_GetCurrentByteIndex(XML_Parser parser)
   return -1;
 }
 
+long XML_GetCurrentByteLimit(XML_Parser parser)
+{
+  if (eventPtr)
+    return parseEndByteIndex - (parseEndPtr - eventEndPtr);
+  return -1;
+}
+
 int XML_GetCurrentLineNumber(XML_Parser parser)
 {
   if (eventPtr) {
@@ -964,7 +1024,7 @@ doContent(XML_Parser parser,
       }
       *eventEndPP = end;
       if (characterDataHandler) {
-	XML_Char c = XML_T('\n');
+	XML_Char c = XML_T(LF);
 	characterDataHandler(handlerArg, &c, 1);
       }
       else if (defaultHandler)
@@ -1064,7 +1124,11 @@ doContent(XML_Parser parser,
 	    entity->open = 0;
 	    if (!context)
 	      return XML_ERROR_NO_MEMORY;
-	    if (!externalEntityRefHandler(parser, context, dtd.base, entity->systemId, entity->publicId))
+	    if (!externalEntityRefHandler(externalEntityRefHandlerArg,
+				          context,
+					  dtd.base,
+					  entity->systemId,
+					  entity->publicId))
 	      return XML_ERROR_EXTERNAL_ENTITY_HANDLING;
 	    poolDiscard(&tempPool);
 	  }
@@ -1115,6 +1179,7 @@ doContent(XML_Parser parser,
 	}
 	++tagLevel;
 	if (startElementHandler) {
+	  int i;
 	  enum XML_Error result;
 	  XML_Char *toPtr;
 	  for (;;) {
@@ -1144,6 +1209,12 @@ doContent(XML_Parser parser,
 	  if (result)
 	    return result;
 	  startElementHandler(handlerArg, tag->name, (const XML_Char **)atts);
+	  for (i = 0; i < attsSize; i += 2) {
+	    if (! ((XML_Char **) atts)[i])
+	      break;
+
+	    (((XML_Char **) atts)[i])[-1] = 0;
+	  }
 	  poolClear(&tempPool);
 	}
 	else {
@@ -1175,8 +1246,16 @@ doContent(XML_Parser parser,
 	if (result)
 	  return result;
 	poolFinish(&tempPool);
-	if (startElementHandler)
+	if (startElementHandler) {
+	  int i;
 	  startElementHandler(handlerArg, name, (const XML_Char **)atts);
+	  for (i = 0; i < attsSize; i += 2) {
+	    if (! ((XML_Char **) atts)[i])
+	      break;
+
+	    (((XML_Char **) atts)[i])[-1] = 0;
+	  }
+	}
 	if (endElementHandler) {
 	  if (startElementHandler)
 	    *eventPP = *eventEndPP;
@@ -1246,7 +1325,7 @@ doContent(XML_Parser parser,
       return XML_ERROR_MISPLACED_XML_PI;
     case XML_TOK_DATA_NEWLINE:
       if (characterDataHandler) {
-	XML_Char c = XML_T('\n');
+	XML_Char c = XML_T(LF);
 	characterDataHandler(handlerArg, &c, 1);
       }
       else if (defaultHandler)
@@ -1255,8 +1334,23 @@ doContent(XML_Parser parser,
     case XML_TOK_CDATA_SECT_OPEN:
       {
 	enum XML_Error result;
-	if (characterDataHandler)
+	if (startCdataSectionHandler)
+  	  startCdataSectionHandler(handlerArg);
+#if 0
+	/* Suppose you doing a transformation on a document that involves
+	   changing only the character data.  You set up a defaultHandler
+	   and a characterDataHandler.  The defaultHandler simply copies
+	   characters through.  The characterDataHandler does the transformation
+	   and writes the characters out escaping them as necessary.  This case
+	   will fail to work if we leave out the following two lines (because &
+	   and < inside CDATA sections will be incorrectly escaped).
+
+	   However, now we have a start/endCdataSectionHandler, so it seems
+	   easier to let the user deal with this. */
+
+	else if (characterDataHandler)
   	  characterDataHandler(handlerArg, dataBuf, 0);
+#endif
 	else if (defaultHandler)
 	  reportDefault(parser, enc, s, next);
 	result = doCdataSection(parser, enc, &next, end, nextPtr);
@@ -1445,14 +1539,14 @@ static enum XML_Error storeAtts(XML_Parser parser, const ENCODING *enc,
 	      return XML_ERROR_NO_MEMORY;
 	  }
           else {
-	    (da->id->name)[-1] = 2;
+	    (da->id->name)[-1] = 6;
 	    nPrefixes++;
   	    appAtts[attIndex++] = da->id->name;
 	    appAtts[attIndex++] = da->value;
 	  }
 	}
 	else {
-	  (da->id->name)[-1] = 1;
+	  (da->id->name)[-1] = 5;
 	  appAtts[attIndex++] = da->id->name;
 	  appAtts[attIndex++] = da->value;
 	}
@@ -1463,7 +1557,7 @@ static enum XML_Error storeAtts(XML_Parser parser, const ENCODING *enc,
   i = 0;
   if (nPrefixes) {
     for (; i < attIndex; i += 2) {
-      if (appAtts[i][-1] == 2) {
+      if ((appAtts[i][-1] & 3) == 2) {
         ATTRIBUTE_ID *id;
         ((XML_Char *)(appAtts[i]))[-1] = 0;
 	id = (ATTRIBUTE_ID *)lookup(&dtd.attributeIds, appAtts[i], 0);
@@ -1491,10 +1585,11 @@ static enum XML_Error storeAtts(XML_Parser parser, const ENCODING *enc,
 	((XML_Char *)(appAtts[i]))[-1] = 0;
     }
   }
-  for (; i < attIndex; i += 2)
-    ((XML_Char *)(appAtts[i]))[-1] = 0;
-  if (!tagNamePtr)
+  if (!tagNamePtr) {
+    for (; i < attIndex; i += 2)
+      ((XML_Char *)(appAtts[i]))[-1] = 0;
     return XML_ERROR_NONE;
+  }
   for (binding = *bindingsPtr; binding; binding = binding->nextTagBinding)
     binding->attId->name[-1] = 0;
   if (elementType->prefix) {
@@ -1614,15 +1709,20 @@ enum XML_Error doCdataSection(XML_Parser parser,
     *eventEndPP = next;
     switch (tok) {
     case XML_TOK_CDATA_SECT_CLOSE:
-      if (characterDataHandler)
+      if (endCdataSectionHandler)
+	endCdataSectionHandler(handlerArg);
+#if 0
+      /* see comment under XML_TOK_CDATA_SECT_OPEN */
+      else if (characterDataHandler)
 	characterDataHandler(handlerArg, dataBuf, 0);
+#endif
       else if (defaultHandler)
 	reportDefault(parser, enc, s, next);
       *startPtr = next;
       return XML_ERROR_NONE;
     case XML_TOK_DATA_NEWLINE:
       if (characterDataHandler) {
-	XML_Char c = XML_T('\n');
+	XML_Char c = XML_T(LF);
 	characterDataHandler(handlerArg, &c, 1);
       }
       else if (defaultHandler)
@@ -2348,7 +2448,7 @@ enum XML_Error storeEntityValue(XML_Parser parser,
     case XML_TOK_DATA_NEWLINE:
       if (pool->end == pool->ptr && !poolGrow(pool))
 	return XML_ERROR_NO_MEMORY;
-      *(pool->ptr)++ = XML_T('\n');
+      *(pool->ptr)++ = XML_T(LF);
       break;
     case XML_TOK_CHAR_REF:
       {
@@ -2392,14 +2492,14 @@ normalizeLines(XML_Char *s)
   for (;; s++) {
     if (*s == XML_T('\0'))
       return;
-    if (*s == XML_T('\r'))
+    if (*s == XML_T(CR))
       break;
   }
   p = s;
   do {
-    if (*s == XML_T('\r')) {
-      *p++ = XML_T('\n');
-      if (*++s == XML_T('\n'))
+    if (*s == XML_T(CR)) {
+      *p++ = XML_T(LF);
+      if (*++s == XML_T(LF))
         s++;
     }
     else
@@ -2727,8 +2827,8 @@ void normalizePublicId(XML_Char *publicId)
   for (s = publicId; *s; s++) {
     switch (*s) {
     case XML_T(' '):
-    case XML_T('\r'):
-    case XML_T('\n'):
+    case XML_T(CR):
+    case XML_T(LF):
       if (p != publicId && p[-1] != XML_T(' '))
 	*p++ = XML_T(' ');
       break;
